#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>
#include <alloca.h>
#include "main.h"
#include "retrieve.h"
#include "misc.h"
#include "wrapper.h"

struct fd_ll *fds = NULL;
struct metar_data_ll *metardata = NULL;
extern int errno;

char *sky_codes[] =
{
	"VV", "SKC", "CLR", "FEW", "SCT", "BKN", "OVC"
};

char *weather_codes[] = 
{
	"BC", "BL", "BR", "DR", "DU", "DZ", "FC", "FG", "FU", "FZ",
	"GR", "GS", "HZ", "IC", "MI", "PE", "PO", "PR", "PY", "RA",
	"SA", "SG", "SH", "SN", "SQ", "SS", "TS", "UP", "VA", "VC"
};

void metar_parse(char **str)
{
	enum token_types tt;

	tt = station;

	while (*str) {
		switch (tt) {
			case station:
				tt = date;
				if (metar_station(*str) == -1)
					continue;
				break;
			case date:
				tt = modifier;
				if (metar_date(*str) == -1)
					continue;
				break;
			case modifier:
				tt = wind;
				if (metar_modifier(*str) == -1)
					continue;
				break;
			case wind:
				tt = visibility;
				if (!strstr(*(str + 1), "SM") && strlen(*(str + 1)) > 3) {	/* variable wind direction >6kt */
					str++;
					if (metar_wind(*(str - 1), *str) == -1)
						continue;
					break;
				} else {
					if (metar_wind(*str, NULL) == -1)
						continue;
					break;
				}
			case visibility:
				tt = runway;
				if ((!strstr(*str, "SM")) && (strlen(*str) < 3)) {
					str++;
					if (metar_visibility(*str, *(str - 1)) == -1)
						continue;
					break;
				} else {
					if (metar_visibility(*str, NULL) == -1)
						continue;
					break;
				}
			case runway:
				tt = weather;
				while (metar_runway(*str) != -1)
					str++;
				continue;
			case weather:
				tt = sky;
				while (metar_weather(*str) != -1)
					str++;
				continue;
			case sky:
				tt = temp;
				while (metar_sky(*str) != -1)
					str++;
				continue;
			case temp:
				tt = altimeter;
				if (metar_temp(*str) == -1)
					continue;
				break;
			case altimeter:
				tt = remark;
				if (metar_altimeter(*str) == -1)
					continue;
				break;
			case remark: /* gotta do this some time */
				tt = finished;
/*
				{
					int c;

					for (c = 0; *(str + c); c++)
						printf("RMK: %s\n", *(str + c));
				}
*/
				break;
			case finished:
				return;
			default:
				printf("default: %s\n", *str);
		}
		str++;
	}
}

int metar_station(char *data)
{
	int i;

	if (!data[0] || !isalpha(data[0]) || (strlen(data) != 4))
		return -1;

	for (i = 1; data[i]; i++) {
		if (!isalnum(data[i]))
			return -1;
	}

	printf("Station: %s\n", data);

	return 0;
}

int metar_date(char *data)
{
	int i;
	int day, hour, min;
	char buf[3];

	if (!data[6] || (data[6] != 'Z'))
		return -1;

	for (i = 0; data[i] && (i < 6); i++) {
		if (!isdigit(data[i]))
			return -1;
	}

	buf[2] = 0;
	buf[0] = data[0];
	buf[1] = data[1];
	day = atoi(buf);
	buf[0] = data[2];
	buf[1] = data[3];
	hour = atoi(buf);
	buf[0] = data[4];
	buf[1] = data[5];
	min = atoi(buf);

	printf("Time of report: %d day of this month at %02d:%02d\n", day, hour, min);

	return 0;
}

int metar_modifier(char *data)
{
	char modifier[5];

	strcpy(modifier, "AUTO");

	if (!match("AUTO", data) && !match("COR", data))
		return -1;

	strcpy(modifier, data);

	printf("Report type: %s\n", modifier);

	return 0;
}

int metar_wind(char *data, char *vwd)
{
	int direction, speed, gust, vwd_low, vwd_high;
	char buf[4];

	direction = speed = gust = vwd_low = vwd_high = 0;

	if (strlen(data) < 7) {
		printf("metar_wind: invalid length for \"%s\"\n", data);
		return -1;
	}
	if (!strstr(data, "KT")) {
		printf("metar_wind: unknown suffix on \"%s\"\n", data);
		return -1;
	}
	buf[3] = 0;
	strncpy(buf, data, 3);

	if (match("VRB", buf))
		direction = -1;
	else
		direction = atoi(buf);

	if (data[7] && data[5] != 'G') {
		strncpy(buf, data + 3, 3);
		data += 6;
	} else {
		buf[0] = '0';
		strncpy(buf + 1, data + 3, 2);
		data += 5;
	}
	speed = atoi(buf);

	if (*data == 'G') {
		if (data[5])
			strncpy(buf, data + 1, 3);
		else {
			buf[0] = '0';
			strncpy(buf + 1, data + 1, 2);
		}
		gust = atoi(buf);
	}
	if (vwd) {
		if (strlen(vwd) != 7) {
			printf("metar_wind: unknown VWD string \"%s\"\n", vwd);
			return -1;
		}
		strncpy(buf, vwd, 3);
		vwd_low = atoi(buf);
		strncpy(buf, vwd + 4, 3);
		vwd_high = atoi(buf);
	}
	printf("Wind: %d kt (%.1f mph) heading %d", speed, kttomph(speed), direction);
	if (vwd_low || vwd_high)
		printf(", varying in direction from %d deg to %d deg", vwd_low, vwd_high);
	if (gust)
		printf(", with gusts up to %d kt (%.1f mph)", gust, kttomph(gust));
	printf("\n");

	return 0;
}

int metar_visibility(char *data, char *extra)
{
	char less_than = 0;
	float visibility;
	char buf[3];
	int t1, t2;

	if (!strstr(data, "SM"))
		return -1;

	if (*data == 'M') {
		less_than = 1;
		data++;
	}
	if (strchr(data, '/')) {
		split(buf, data, '/');
		t1 = atoi(buf);
		split(buf, data, 'S');
		t2 = atoi(buf);
		visibility = t1 * 1.0 / t2;
	} else {
		split(buf, data, 'S');
		visibility = atof(buf);
	}

	if (extra)
		visibility += atoi(extra) * 1.0;

	printf("Visibility: %s%.4f miles\n", less_than ? "less than " : "", visibility);
	return 0;
}

int metar_runway(char *data)
{
	int len = strlen(data);

	if (len < 2)
		return -1;
	if (!strchr(data, '/') || ((data[len - 2] != 'F') && (data[len - 1] != 'T')))
		return -1;

	printf("Runway Information: currently unimplemented (%s)\n", data);
	return 0;
}

int metar_weather(char *data)
{
	char buf[3];
	int i;
	
	if (*data == '+' || *data == '-')
		data++;

	buf[2] = 0;

	for (i = 0; data[i]; i += 2) {
		strncpy(buf, data + i, 2);
		if (!is_array_item(weather_codes, arraysize(weather_codes), buf))
			return -1;
	}
	printf("Current Weather Code: %s\n", data);

	return 0;
}

int metar_sky(char *data)
{
	int i, count, height;

	count = height = 0;

	for (i = 0; i < arraysize(sky_codes); i++) {
		if (strstr(data, sky_codes[i])) {
			count++;
			break;
		}
	}

	if (!count)
		return -1;

	if (!i) {
		data += 2;
		height = atoi(data);
	} else if (i > 2) {
		data += 3;
		if (match(data, "///"))
			height = -50;
		else
			height = atoi(data);
	}
	printf("Sky: %s at %d feet\n", sky_codes[i], height * 100);
	return 0;
}

int metar_temp(char *data)
{
	char buf[4];
	int dewpoint, temperature;
	
	temperature = dewpoint = -5000;

	if (!strchr(data, '/') || strlen(data) > 7) 
		return -1;
		
	split(buf, data, '/');
	
	if (buf[0] == 'M')
		temperature = -atoi(buf + 1);
	else
		temperature = atoi(buf);	

	if (data[0]) {
		if (data[0] == 'M')
			dewpoint = -atoi(data + 1);
		else
			dewpoint = atoi(data);
	}
	
	printf("Temperature: %d (%.1fF)\n", temperature, ctof(temperature));
	printf("Dewpoint: %d (%.1fF)\n", dewpoint, ctof(dewpoint));
	return 0;
}

int metar_altimeter(char *data)
{
	float altimeter;

	if (data[0] != 'A' || strlen(data) != 5)
		return -1;
	
	altimeter = atof(data + 1) / 100;

	printf("Altimeter: %.2f inches Hg\n", altimeter);
	return 0;
}

int main(int argc, char **argv)
{
	char **code, *data, *request, *path;
	struct fd_ll *f, *next;
	struct metar_data_ll *m;
	fd_set fdset;
	struct timeval tv;

	if (argc != 2) {
		printf("usage: %s <METAR station>\n", argv[0]);
		return 1;
	}

	if (metar_station(argv[1]) != 0) {
		printf("%s: illegal station name \"%s.\"\n", argv[0], argv[1]);
		return 1;
	}

	data = NULL;
	path = HTML_PATH;	
	request = alloca(strlen(path) + 20);
	
	snprintf(request, strlen(path) + 20, "GET %s%s HTTP/1.0\n\n", path, argv[1]);

	metar_request(HTML_SOURCE, HTML_PORT, argv[1], request);

	for (;;) {
		FD_ZERO(&fdset);

		f = fds;
		while (f) {
			FD_SET(f->fd, &fdset);
			f = f->next;
		}

		tv.tv_sec = 1;
		tv.tv_usec = 0;

		if (select(HIGH_FD, &fdset, NULL, NULL, &tv) < 0) {
			if (errno != EINTR)
				printf("WARNING: select error (%s)", strerror(errno));
			continue;
		}

		f = fds;
		while (f) {
			next = f->next;
			if (FD_ISSET(f->fd, &fdset)) {
				f->handle(f->fd);
				m = find_metar_data(f->id);

				if (!f->parse_state) {
					m->string = NULL;
				}
				if (f->parse_state == 2) {
					printf("Observation: %s\n", m->string);
				}
				if (f->parse_state == 3) {
					if (m->string && strlen(m->string)) {
						code = separate_data(m->string);
						metar_parse(code);
						for (; *code; code++)
							my_free(*code);
					} else {
						printf("No METAR data available for station \"%s.\"\n", f->id);
						del_metar_data(f->id);
					}
					del_fd(f->fd);
				}
			}
			f = next;
		}
		if (!fds)
			break;
	}

	/* one to take entire tree down; one to take just the limb off the
	   roof; leave wood here but cut into fireplace lengths. */

	del_all_metar_data();

	return 0;
}
